(define (ast-type x)
  (define (builtin? x)
    (case x
      ('+ #t)
      ('- #t)
      ('* #t)
      ('! #t)
      ('= #t)
      ('bool->int #t)
      ('print #t)
      (else #f)))
  (cond
   ((list? x)
    (case (car x)
      ('if 'if)
      ('let 'let)
      ('lambda 'lambda)
      ('closure 'closure) ; only available in codegen
      ('static-string 'static-string) ; only available in codegen
      (else 'app)))
   ((builtin? x) 'builtin)
   ((symbol? x) 'var)
   ((integer? x) 'int-literal)
   ((boolean? x) 'bool-literal)
   ((string? x) 'string-literal)))

(define (ast-traverse f x)
  (case (ast-type x)
    ('let `(let ,(map (lambda (x) (list (car x) (f (cadr x))))
		      (let-bindings x))
	     ,@(map f (let-body x))))
    ('app (map f x))
    ('lambda `(lambda ,(lambda-args x) ,(f (lambda-body x))))
    ('if `(if ,@(map f (cdr x))))
    (else x)))

(define (ast-collect f x)
  (define (inner y) (ast-collect f y))
  (case (ast-type x)
    ['let (append (f x)
		  (fold-map inner (let-bindings x))
		  (fold-map inner (let-body x)))]
    ['app (append (f x)
		  (fold-map inner x))]
    ['lambda (append (f x)
		     (inner (lambda-body x)))]
    ['if (append (f x)
		 (fold-map inner (cdr x)))]
    [else (f x)]))

(define let-bindings cadr)
(define let-body cddr)

(define (lambda? x)
  (and (list? x) (eq? (car x) 'lambda)))

; for use in normalized form
(define lambda-arg caadr)
; for use elsewhere
(define lambda-args cadr)
(define lambda-body caddr)

; utils
(define (fold-map f x) (fold-left append '() (map f x)))
(define (repeat x n) (if (<= n 0) '()
			 (cons x (repeat x (- n 1)))))


(define-syntax push!
  (syntax-rules ()
    ((_ s x) (set! s (cons x s)))))

(define-syntax pop!
  (syntax-rules ()
    ((_ s) (let ([x (car s)])
	     (set! s (cdr s))
	     x))))
